#!/usr/bin/env python3

from os import environ
from sys import exit

import dotenv
import hydra
from omegaconf import DictConfig

dotenv.load_dotenv(override=True)

hydra_config_path = environ.get('DEEPMOONCONFIGPATH', "..")


@hydra.main(config_path=f"{hydra_config_path}/configs/",
            config_name="train.yaml")
def train(config: DictConfig) -> None:

    # Imports can be nested inside @hydra.main to optimize tab completion
    # https://github.com/facebookresearch/hydra/issues/934
    from torchmoon.logger import extras
    from torchmoon.training import training

    extras(config)

    return training(config)

exit(train())



from typing import List
from omegaconf import DictConfig

from pytorch_lightning import (LightningModule, LightningDataModule, Trainer,
                               Callback)
from pytorch_lightning.loggers import (LightningLoggerBase, wandb)


def log_hyperparameters(
    config: DictConfig,
    model: LightningModule,
    datamodule: LightningDataModule,
    trainer: Trainer,
    callbacks: List[Callback],
    logger: List[LightningLoggerBase],
) -> None:
    """Controls which config parts are saved by Lightning loggers.
    Additionaly saves:
    - number of model parameters
    """

    hparams = {}

    # choose which parts of hydra config are saved by loggers
    hparams["trainer"] = config["trainer"]
    hparams["model"] = config["model"]
    hparams["datamodule"] = config["datamodule"]

    if "seed" in config:
        hparams["seed"] = config["seed"]
    if "callbacks" in config:
        hparams["callbacks"] = config["callbacks"]

    # save number of model parameters
    hparams["model/params/total"] = sum(p.numel() for p in model.parameters())
    hparams["model/params/trainable"] = sum(p.numel()
                                            for p in model.parameters()
                                            if p.requires_grad)
    hparams["model/params/non_trainable"] = sum(p.numel()
                                                for p in model.parameters()
                                                if not p.requires_grad)

    # send hparams to all loggers
    trainer.logger.log_hyperparams(hparams)


def finish(
    config: DictConfig,
    model: LightningModule,
    datamodule: LightningDataModule,
    trainer: Trainer,
    callbacks: List[Callback],
    logger: List[LightningLoggerBase],
) -> None:
    """Makes sure everything closed properly."""

    # without this sweeps with wandb logger might crash!
    for lg in logger:
        if isinstance(lg, wandb.WandbLogger):
            import wandb

            wandb.finish()



from typing import (Tuple, Optional)

from math import floor

from torch import Generator
from torch.utils.data import DataLoader, Dataset, random_split
from torchvision import transforms

from pytorch_lightning import LightningDataModule

from craterdata.mooncraterdataset import MoonCraterDataset


class CraterDataModule(LightningDataModule):

    def __init__(self,
                 data_dir: str = "data/",
                 batch_size: int = 256,
                 num_worker: int = 8,
                 image_size: int = 256,
                 training_test_eval_split: Tuple[float, float,
                                                 float] = (0.25, 0.25, 0.75),
                 download: bool = False,
                 pin_memory: bool = False) -> None:

        super().__init__()

        self.save_hyperparameters(logger=False)

        self.transform = transforms.Compose([
            transforms.Resize([self.hparams.image_size, self.hparams.image_size]),
            transforms.ToTensor(),
        ])

        self.data_train: Optional[Dataset] = None
        self.data_val: Optional[Dataset] = None
        self.data_test: Optional[Dataset] = None

        self.crater_dataset = None

    def prepare_data(self):
        if self.crater_dataset is None:
            self.crater_dataset = MoonCraterDataset(
                root=self.hparams.data_dir,
                transform=self.transform,
                target_transform=self.transform,
                download=self.hparams.download)

    def setup(self, stage: Optional[str] = None):

        train_val_test_split = [
            floor(value * len(self.crater_dataset))
            for value in self.hparams.training_test_eval_split
        ]

        if self.data_train is None and self.data_val is None and self.data_test is None:
            self.data_train, self.data_val, self.data_test = random_split(
                dataset=self.crater_dataset,
                lengths=train_val_test_split,
                generator=Generator().manual_seed(42))

    def train_dataloader(self):
        return DataLoader(dataset=self.data_train,
                          batch_size=self.hparams.batch_size,
                          num_workers=self.hparams.num_worker,
                          pin_memory=self.hparams.pin_memory,
                          shuffle=True)

    def val_dataloader(self):
        return DataLoader(dataset=self.data_val,
                          batch_size=self.hparams.batch_size,
                          num_workers=self.hparams.num_worker,
                          pin_memory=self.hparams.pin_memory,
                          shuffle=False)

    def test_dataloader(self):
        return DataLoader(dataset=self.data_test,
                          batch_size=self.hparams.batch_size,
                          num_workers=self.hparams.num_worker,
                          pin_memory=self.hparams.pin_memory,
                          shuffle=False)



from torch.nn import (Module, Conv2d, Sequential)

from torchmoon.torch.activations import Activation
from torchmoon.torch.util import ContBatchNorm2d

class LUConv(Module):
    def __init__(self, nchan, relu):
        super().__init__()
        self.relu1 = Activation(relu, nchan)
        self.conv1 = Conv2d(nchan, nchan, kernel_size=5, padding=2)
        self.bn1 = ContBatchNorm2d(nchan)

    def forward(self, idata):
        idata = self.relu1(self.bn1(self.conv1(idata)))
        return idata


def make_Conv(nchan, depth, relu):
    layers = list()
    for _ in range(depth):
        layers.append(LUConv(nchan, relu))
    return Sequential(*layers)


class ConvBlock(Module):
    def __init__(self, ichan, ochan, relu):
        super().__init__()
        self.conv1 = Conv2d(ichan, ochan, kernel_size=5, padding=2)
        self.activation = Activation(relu, ochan)

    def forward(self, idata):
        idata = self.activation(self.conv1(idata))
        return idata



from typing import (Any, List)

from torch.nn import (Module, Conv2d, Dropout2d, ConvTranspose2d)
from torch import (cat, add)
from torch.nn import BCEWithLogitsLoss
from torch.optim import Adam
import pytorch_lightning as pl

from torchmetrics import (MaxMetric, StructuralSimilarityIndexMeasure)

from torchmoon.torch.activations import Activation
from torchmoon.torch.util import (passthrough, ContBatchNorm2d)
from torchmoon.torch.conv import make_Conv


class InputTransition(Module):

    def __init__(self, relu):
        super().__init__()

        self.conv1 = Conv2d(1, 16, kernel_size=5, padding=2)
        self.bn1 = ContBatchNorm2d(16)
        self.relu1 = Activation(relu, 16)

    def forward(self, idata):
        outdata = self.relu1(self.bn1(self.conv1(idata)))

        return outdata


#start downpath
class DownTransition(Module):

    def __init__(self, inchan, nConvs, relu, dropout=0):
        super().__init__()

        outchan = 2 * inchan
        self.down_conv = Conv2d(inchan, outchan, kernel_size=2, stride=2)
        self.bn1 = ContBatchNorm2d(outchan)
        self.do1 = passthrough
        self.relu1 = Activation(relu, outchan)
        self.relu2 = Activation(relu, outchan)

        if dropout > 0:
            self.do1 = Dropout2d(dropout)

        self.ops = make_Conv(outchan, nConvs, relu)

    def forward(self, idata):
        down = self.relu1(self.bn1(self.down_conv(idata)))
        out = self.do1(down)
        out = self.ops(out)
        out = self.relu2(add(out, down))
        return out


class UpTransition(Module):

    def __init__(self, inchan, outchan, nConvs, relu, dropout=0):
        super().__init__()

        self.up_conv = ConvTranspose2d(inchan,
                                       outchan // 2,
                                       kernel_size=2,
                                       stride=2)
        self.bn1 = ContBatchNorm2d(outchan // 2)
        self.do1 = passthrough
        self.do2 = Dropout2d(dropout)
        self.relu1 = Activation(relu, outchan // 2)
        self.relu2 = Activation(relu, outchan)

        if dropout > 0:
            self.do1 = Dropout2d(dropout)

        self.ops = make_Conv(outchan, nConvs, relu)

    def forward(self, idata, skipx):
        up = self.do1(idata)
        skipxdo = self.do2(skipx)
        out = self.relu1(self.bn1(self.up_conv(up)))
        xcat = cat((out, skipxdo), 1)
        out = self.ops(xcat)
        out = self.relu2(add(out, xcat))
        return out


class OutTransition(Module):

    def __init__(self, inchan, relu):
        super().__init__()

        self.conv1 = Conv2d(inchan, 2, kernel_size=5, padding=2)
        self.bn1 = ContBatchNorm2d(2)
        self.conv2 = Conv2d(2, 2, kernel_size=1)
        self.relu1 = Activation(relu, 2)
        self.conv3 = Conv2d(2, 1, kernel_size=1)

    def forward(self, idata):
        out = self.relu1(self.bn1(self.conv1(idata)))
        out = self.relu1(self.conv2(out))
        out = self.conv3(out)
        return out


class Crater_VNet(pl.LightningModule):

    def __init__(self, activation="relu", dropout=.15, lr=0.02):
        super().__init__()
        self.save_hyperparameters(logger=False)

        self.criterion = BCEWithLogitsLoss()

        # use separate metric instance for train, val and test step
        # to ensure a proper reduction over the epoch
        self.train_acc = StructuralSimilarityIndexMeasure()
        self.val_acc = StructuralSimilarityIndexMeasure()
        self.test_acc = StructuralSimilarityIndexMeasure()

        # for logging best so far validation accuracy
        self.val_acc_best = MaxMetric()

        self.in_tr = InputTransition(self.hparams.activation)
        self.down_32 = DownTransition(16, 1, self.hparams.activation)
        self.down_64 = DownTransition(32, 1, self.hparams.activation)
        self.down_128 = DownTransition(64, 2, self.hparams.activation,
                                       self.hparams.dropout)
        self.down_256 = DownTransition(128, 2, self.hparams.activation,
                                       self.hparams.dropout)
        self.up_256 = UpTransition(256, 256, 2, self.hparams.activation,
                                   self.hparams.dropout)
        self.up_128 = UpTransition(256, 128, 2, self.hparams.activation,
                                   self.hparams.dropout)
        self.up_64 = UpTransition(128, 64, 1, self.hparams.activation)
        self.up_32 = UpTransition(64, 32, 1, self.hparams.activation)
        self.out_tr = OutTransition(32, self.hparams.activation)

    def forward(self, idata):
        out16 = self.in_tr(idata)
        out32 = self.down_32(out16)
        out64 = self.down_64(out32)
        out128 = self.down_128(out64)
        out256 = self.down_256(out128)

        out = self.up_256(out256, out128)
        out = self.up_128(out, out64)
        out = self.up_64(out, out32)
        out = self.up_32(out, out16)
        out = self.out_tr(out)

        return out

    def configure_optimizers(self):
        optimizer = Adam(self.parameters(), lr=self.hparams.lr)
        return optimizer

    def step(self, batch: Any):
        x, y, _ = batch

        y_hat = self.forward(x)
        loss = self.criterion(y_hat, y)

        return loss, y_hat, y

    def training_step(self, train_batch: Any, batch_idx: int):
        # data to device
        loss, preds, targets = self.step(train_batch)

        acc = self.train_acc(preds, targets)
        self.log("train/loss",
                 loss,
                 on_step=False,
                 on_epoch=True,
                 prog_bar=False)
        self.log("train/acc", acc, on_step=False, on_epoch=True, prog_bar=True)

        return {"loss": loss, "preds": preds, "targets": targets}

    def validation_step(self, val_batch: Any, batch_idx: int):
        loss, preds, targets = self.step(val_batch)

        # log val metrics
        acc = self.val_acc(preds, targets)
        self.log("val/loss",
                 loss,
                 on_step=False,
                 on_epoch=True,
                 prog_bar=False)
        self.log("val/acc", acc, on_step=False, on_epoch=True, prog_bar=True)

        return {"loss": loss, "preds": preds, "targets": targets}

    def validation_epoch_end(self, outputs: List[Any]):
        acc = self.val_acc.compute()  # get val accuracy from current epoch
        self.val_acc_best.update(acc)
        self.log("val/acc_best",
                 self.val_acc_best.compute(),
                 on_epoch=True,
                 prog_bar=True)

    def test_step(self, batch: Any, batch_idx: int):
        loss, preds, targets = self.step(batch)

        # log test metrics
        acc = self.test_acc(preds, targets)
        self.log("test/loss", loss, on_step=False, on_epoch=True)
        self.log("test/acc", acc, on_step=False, on_epoch=True)

        return {"loss": loss, "preds": preds, "targets": targets}

    def on_epoch_end(self):
        # reset metrics at the end of every epoch
        self.train_acc.reset()
        self.test_acc.reset()
        self.val_acc.reset()



from torch.nn import (ReLU, PReLU, ELU)

def Activation(activation, nchan):
    if "elu" == activation:
        return ELU(inplace=True)
    elif "prelu" == activation:
        return PReLU(nchan)
    else:
        return ReLU(inplace=True)



from torch import cat
from torch.nn.modules.batchnorm import _BatchNorm
from torch.nn.functional import batch_norm

class ContBatchNorm2d(_BatchNorm):
    def _check_input_dim(self, idata):
        if idata.dim() != 4:
            raise ValueError(f"expected 4d input. I got {idata.dim()}")

    def forward(self, idata):
        self._check_input_dim(idata)
        return batch_norm(idata, self.running_mean, self.running_var,
                          self.weight, self.bias, True, self.momentum,
                          self.eps)

def merge(layers, cat_axis=None):
    return cat(layers, dim=cat_axis)

def passthrough(x, **kwargs):
    return x



from typing import Any, List

from torch.nn import (Conv2d, Sequential, Dropout2d, Upsample)
from torch.nn.init import xavier_uniform
from torch.nn.modules.activation import Sigmoid
from torch.nn import BCEWithLogitsLoss
from torch.nn.modules.pooling import MaxPool2d
from torch.optim import Adam
import pytorch_lightning as pl

from torchmetrics import (MaxMetric, StructuralSimilarityIndexMeasure)

from torchmoon.torch.activations import Activation
from torchmoon.torch.util import merge


class DeepMoon(pl.LightningModule):

    def __init__(self,
                 number_of_filter,
                 filter_length,
                 lmbda=1e-6,
                 dim=256,
                 activation="relu",
                 dropout=.15,
                 lr=1e-5):
        super().__init__()

        # this line allows to access init params with 'self.hparams' attribute
        # it also ensures init params will be stored in ckpt
        self.save_hyperparameters(logger=False)

        self.criterion = BCEWithLogitsLoss()

        # use separate metric instance for train, val and test step
        # to ensure a proper reduction over the epoch
        self.train_acc = StructuralSimilarityIndexMeasure()
        self.val_acc = StructuralSimilarityIndexMeasure()
        self.test_acc = StructuralSimilarityIndexMeasure()
        
        # for logging best so far validation accuracy
        self.val_acc_best = MaxMetric()

        self.down_0 = Sequential(
            Conv2d(
                in_channels=1,  # gray image
                out_channels=self.hparams.number_of_filter,
                kernel_size=(self.hparams.filter_length,
                             self.hparams.filter_length),
                padding=1),
            Activation(self.hparams.activation, self.hparams.number_of_filter),
            Conv2d(in_channels=self.hparams.number_of_filter,
                   out_channels=self.hparams.number_of_filter,
                   kernel_size=(self.hparams.filter_length,
                                self.hparams.filter_length),
                   padding=1),
            Activation(self.hparams.activation, self.hparams.number_of_filter))

        number_of_filter_2 = self.hparams.number_of_filter * 2
        self.down_1 = Sequential(
            Conv2d(in_channels=self.hparams.number_of_filter,
                   out_channels=number_of_filter_2,
                   kernel_size=self.hparams.filter_length,
                   padding=1),
            Activation(self.hparams.activation, number_of_filter_2),
            Conv2d(in_channels=number_of_filter_2,
                   out_channels=number_of_filter_2,
                   kernel_size=self.hparams.filter_length,
                   padding=1),
            Activation(self.hparams.activation, number_of_filter_2))

        number_of_filter_4 = self.hparams.number_of_filter * 4
        self.down_2 = Sequential(
            Conv2d(in_channels=number_of_filter_2,
                   out_channels=number_of_filter_4,
                   kernel_size=self.hparams.filter_length,
                   padding=1),
            Activation(self.hparams.activation, number_of_filter_4),
            Conv2d(in_channels=number_of_filter_4,
                   out_channels=number_of_filter_4,
                   kernel_size=self.hparams.filter_length,
                   padding=1),
            Activation(self.hparams.activation, number_of_filter_4),
        )

        self.base = Sequential(
            Conv2d(in_channels=number_of_filter_4,
                   out_channels=number_of_filter_4,
                   kernel_size=self.hparams.filter_length,
                   padding=1),
            Activation(self.hparams.activation, self.hparams.number_of_filter),
            Conv2d(in_channels=number_of_filter_4,
                   out_channels=number_of_filter_4,
                   kernel_size=self.hparams.filter_length,
                   padding=1),
            Activation(self.hparams.activation, number_of_filter_4))

        self.down_0.apply(self.init_weights)
        self.down_1.apply(self.init_weights)
        self.down_2.apply(self.init_weights)
        self.base.apply(self.init_weights)

        self.up_1 = Sequential(
            Conv2d(in_channels=self.hparams.number_of_filter * 8,
                   out_channels=number_of_filter_2,
                   kernel_size=self.hparams.filter_length,
                   padding=1),
            Activation(self.hparams.activation, number_of_filter_2),
            Conv2d(in_channels=number_of_filter_2,
                   out_channels=number_of_filter_2,
                   kernel_size=self.hparams.filter_length,
                   padding=1),
            Activation(self.hparams.activation, number_of_filter_2))

        self.up_2 = Sequential(
            Conv2d(in_channels=number_of_filter_4,
                   out_channels=self.hparams.number_of_filter,
                   kernel_size=self.hparams.filter_length,
                   padding=1),
            Activation(self.hparams.activation, self.hparams.number_of_filter),
            Conv2d(in_channels=self.hparams.number_of_filter,
                   out_channels=self.hparams.number_of_filter,
                   kernel_size=self.hparams.filter_length,
                   padding=1),
            Activation(self.hparams.activation, self.hparams.number_of_filter))

        self.up_3 = Sequential(
            Conv2d(in_channels=number_of_filter_2,
                   out_channels=self.hparams.number_of_filter,
                   kernel_size=self.hparams.filter_length,
                   padding=1),
            Activation(self.hparams.activation, self.hparams.number_of_filter),
            Conv2d(in_channels=self.hparams.number_of_filter,
                   out_channels=self.hparams.number_of_filter,
                   kernel_size=self.hparams.filter_length,
                   padding=1),
            Activation(self.hparams.activation, self.hparams.number_of_filter))

        self.up_1.apply(self.init_weights)
        self.up_2.apply(self.init_weights)
        self.up_3.apply(self.init_weights)

        self.out_conv = Conv2d(in_channels=self.hparams.number_of_filter,
                               out_channels=1,
                               kernel_size=1)

        self.out_conv.apply(self.init_weights)

    def forward(self, idata):
        # down
        d0 = self.down_0(idata)
        max_1 = MaxPool2d(kernel_size=(2, 2), stride=(2, 2))(d0)
        d1 = self.down_1(max_1)
        max_2 = MaxPool2d(kernel_size=(2, 2), stride=(2, 2))(d1)
        d2 = self.down_2(max_2)
        max_3 = MaxPool2d(kernel_size=(2, 2), stride=(2, 2))(d2)

        # base
        u = self.base(max_3)
        u = Upsample(scale_factor=2, mode='nearest')(u)

        # upscale
        u = merge(layers=(d2, u), cat_axis=1)
        u = self.dropout_reg(u)
        u = self.up_1(u)
        u = Upsample(scale_factor=2, mode='nearest')(u)

        u = merge(layers=(d1, u), cat_axis=1)
        u = self.dropout_reg(u)
        u = self.up_2(u)
        u = Upsample(scale_factor=2, mode='nearest')(u)

        u = merge(layers=(d0, u), cat_axis=1)
        u = self.dropout_reg(u)
        u = self.up_3(u)

        u = self.out_conv(u)
        u = Sigmoid()(u)

        return u

    def dropout_reg(self, u):
        if self.hparams.dropout is not None and self.hparams.dropout > 0:
            return Dropout2d(p=self.hparams.dropout, inplace=True)(u)
        return u

    def init_weights(self, m):
        if isinstance(m, Conv2d):
            xavier_uniform(m.weight)
            #m.bias.data.fill_(self.lmbda)

    def configure_optimizers(self):
        optimizer = Adam(self.parameters(),
                         lr=self.hparams.lr,
                         weight_decay=self.hparams.lmbda)
        return optimizer

    def step(self, batch: Any):
        x, y, _ = batch

        y_hat = self.forward(x)
        loss = self.criterion(y_hat, y)

        return loss, y_hat, y

    def training_step(self, train_batch: Any, batch_idx: int):
        # data to device
        loss, preds, targets = self.step(train_batch)

        acc = self.train_acc(preds, targets)
        self.log("train/loss",
                 loss,
                 on_step=False,
                 on_epoch=True,
                 prog_bar=False)
        self.log("train/acc", acc, on_step=False, on_epoch=True, prog_bar=True)

        return {"loss": loss, "preds": preds, "targets": targets}

    def validation_step(self, val_batch: Any, batch_idx: int):
        loss, preds, targets = self.step(val_batch)

        # log val metrics
        acc = self.val_acc(preds, targets)
        self.log("val/loss",
                 loss,
                 on_step=False,
                 on_epoch=True,
                 prog_bar=False)
        self.log("val/acc", acc, on_step=False, on_epoch=True, prog_bar=True)

        return {"loss": loss, "preds": preds, "targets": targets}

    def validation_epoch_end(self, outputs: List[Any]):
        acc = self.val_acc.compute()  # get val accuracy from current epoch
        self.val_acc_best.update(acc)
        self.log("val/acc_best",
                 self.val_acc_best.compute(),
                 on_epoch=True,
                 prog_bar=True)

    def test_step(self, batch: Any, batch_idx: int):
        loss, preds, targets = self.step(batch)

        # log test metrics
        acc = self.test_acc(preds, targets)
        self.log("test/loss", loss, on_step=False, on_epoch=True)
        self.log("test/acc", acc, on_step=False, on_epoch=True)

        return {"loss": loss, "preds": preds, "targets": targets}

    def on_epoch_end(self):
        # reset metrics at the end of every epoch
        self.train_acc.reset()
        self.test_acc.reset()
        self.val_acc.reset()



#!/usr/bin/env python

from os import walk
from os.path import join

from setuptools import find_packages, setup

from torchmoon.__version__ import VERSION
from torchmoon import __author__, __email__

def package_files(directory):
    paths = []
    for (path, _, filenames) in walk(directory):
        for filename in filenames:
            paths.append(join(path, filename))
    return paths

extra_files = package_files('configs')

setup(name="torchmoon",
      version=VERSION,
      description="Crater detection",
      author=__author__,
      author_email=__email__,
      url='https://github.com/afeldman/TorchMoon',
      packages=find_packages(),
      package_data={'': extra_files},)
